@@ -466,11 +466,14 @@ ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
 			q++;
 		}
 
-		len = ntfs_mbstoucs(p, &unicode, NTFS_MAX_NAME_LEN);
+		len = ntfs_mbstoucs(p, &unicode, MAX_PATH);
 		if (len < 0) {
 			ntfs_log_debug("Couldn't convert name to Unicode: %s.\n", p);
 			err = errno;
 			goto close;
+		} else if (len > NTFS_MAX_NAME_LEN) {
+			err = ENAMETOOLONG;
+			goto close;
 		}
 
 		inum = ntfs_inode_lookup_by_name(ni, unicode, len);
@@ -482,7 +485,10 @@ ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
 		}
 
 		if (ni != parent)
-			ntfs_inode_close(ni);
+			if (ntfs_inode_close(ni)) {
+				err = errno;
+				goto out;
+			}
 
 		inum = MREF(inum);
 		ni = ntfs_inode_open(vol, inum);
@@ -502,7 +508,9 @@ ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
 	ni = NULL;
 close:
 	if (ni && (ni != parent))
-		ntfs_inode_close(ni);
+		if (ntfs_inode_close(ni) && !err)
+			err = errno;
+out:
 	free(ascii);
 	free(unicode);
 	if (err)
@@ -570,6 +578,8 @@ static int ntfs_filldir(ntfs_inode *dir_ni, s64 *pos, u8 ivcn_bits,
 		return 0;
 	if (ie->key.file_name.file_attributes & FILE_ATTR_I30_INDEX_PRESENT)
 		dt_type = NTFS_DT_DIR;
+	else if (fn->file_attributes & FILE_ATTR_SYSTEM)
+		dt_type = NTFS_DT_UNKNOWN;
 	else
 		dt_type = NTFS_DT_REG;
 	return filldir(dirent, fn->file_name, fn->file_name_length,
@@ -716,7 +726,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 				le16_to_cpu(dir_ni->mrec->sequence_number)),
 				NTFS_DT_DIR);
 		if (rc)
-			goto done;
+			goto err_out;
 		++*pos;
 	}
 	if (*pos == 1) {
@@ -731,7 +741,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 		rc = filldir(dirent, dotdot, 2, FILE_NAME_POSIX, *pos,
 				parent_mref, NTFS_DT_DIR);
 		if (rc)
-			goto done;
+			goto err_out;
 		++*pos;
 	}
 
@@ -744,8 +754,8 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 	/* Find the index root attribute in the mft record. */
 	if (ntfs_attr_lookup(AT_INDEX_ROOT, NTFS_INDEX_I30, 4, CASE_SENSITIVE, 0, NULL,
 			0, ctx)) {
-		ntfs_log_debug("Index root attribute missing in directory inode "
-				"0x%llx.\n", (unsigned long long)dir_ni->mft_no);
+		ntfs_log_perror("Index root attribute missing in directory inode "
+				"0x%llx", (unsigned long long)dir_ni->mft_no);
 		goto dir_err_out;
 	}
 	/* Get to the index root value. */
@@ -756,7 +766,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 	index_block_size = le32_to_cpu(ir->index_block_size);
 	if (index_block_size < NTFS_BLOCK_SIZE ||
 			index_block_size & (index_block_size - 1)) {
-		ntfs_log_debug("Index block size %u is invalid.\n",
+		ntfs_log_error("Index block size %u is invalid.\n",
 				(unsigned)index_block_size);
 		goto dir_err_out;
 	}
@@ -808,7 +818,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 		if (rc) {
 			ntfs_attr_put_search_ctx(ctx);
 			ctx = NULL;
-			goto done;
+			goto err_out;
 		}
 	}
 	ntfs_attr_put_search_ctx(ctx);
@@ -842,7 +852,7 @@ skip_index_root:
 
 	bmp_pos = ia_pos >> index_block_size_bits;
 	if (bmp_pos >> 3 >= bmp_na->data_size) {
-		ntfs_log_debug("Current index position exceeds index bitmap "
+		ntfs_log_error("Current index position exceeds index bitmap "
 				"size.\n");
 		goto dir_err_out;
 	}
@@ -873,6 +883,7 @@ find_next_index_buffer:
 		if (bmp_buf_pos >> 3 < bmp_buf_size)
 			continue;
 		/* Read next chunk from the index bitmap. */
+		bmp_buf_pos = 0;
 		if ((bmp_pos >> 3) + bmp_buf_size > bmp_na->data_size)
 			bmp_buf_size = bmp_na->data_size - (bmp_pos >> 3);
 		br = ntfs_attr_pread(bmp_na, bmp_pos >> 3, bmp_buf_size, bmp);
@@ -899,7 +910,7 @@ find_next_index_buffer:
 	ia_start = ia_pos & ~(s64)(index_block_size - 1);
 	if (sle64_to_cpu(ia->index_block_vcn) != ia_start >>
 			index_vcn_size_bits) {
-		ntfs_log_debug("Actual VCN (0x%llx) of index buffer is different "
+		ntfs_log_error("Actual VCN (0x%llx) of index buffer is different "
 				"from expected VCN (0x%llx) in inode 0x%llx.\n",
 				(long long)sle64_to_cpu(ia->index_block_vcn),
 				(long long)ia_start >> index_vcn_size_bits,
@@ -907,7 +918,7 @@ find_next_index_buffer:
 		goto dir_err_out;
 	}
 	if (le32_to_cpu(ia->index.allocated_size) + 0x18 != index_block_size) {
-		ntfs_log_debug("Index buffer (VCN 0x%llx) of directory inode 0x%llx "
+		ntfs_log_error("Index buffer (VCN 0x%llx) of directory inode 0x%llx "
 				"has a size (%u) differing from the directory "
 				"specified size (%u).\n", (long long)ia_start >>
 				index_vcn_size_bits,
@@ -918,7 +929,7 @@ find_next_index_buffer:
 	}
 	index_end = (u8*)&ia->index + le32_to_cpu(ia->index.index_length);
 	if (index_end > (u8*)ia + index_block_size) {
-		ntfs_log_debug("Size of index buffer (VCN 0x%llx) of directory inode "
+		ntfs_log_error("Size of index buffer (VCN 0x%llx) of directory inode "
 				"0x%llx exceeds maximum size.\n",
 				(long long)ia_start >> index_vcn_size_bits,
 				(unsigned long long)dir_ni->mft_no);
@@ -940,7 +951,7 @@ find_next_index_buffer:
 				sizeof(INDEX_ENTRY_HEADER) > index_end ||
 				(u8*)ie + le16_to_cpu(ie->key_length) >
 				index_end) {
-			ntfs_log_debug("Index entry out of bounds in directory inode "
+			ntfs_log_error("Index entry out of bounds in directory inode "
 				"0x%llx.\n", (unsigned long long)dir_ni->mft_no);
 			goto dir_err_out;
 		}
@@ -957,7 +968,7 @@ find_next_index_buffer:
 		rc = ntfs_filldir(dir_ni, pos, index_vcn_size_bits,
 				INDEX_TYPE_ALLOCATION, ia, ie, dirent, filldir);
 		if (rc)
-			goto done;
+			goto err_out;
 	}
 	goto find_next_index_buffer;
 EOD:
@@ -970,13 +981,7 @@ done:
 		ntfs_attr_close(bmp_na);
 	if (ia_na)
 		ntfs_attr_close(ia_na);
-#ifdef DEBUG
-	if (!rc)
-		ntfs_log_debug("EOD, *pos 0x%llx, returning 0.\n", (long long)*pos);
-	else
-		ntfs_log_debug("filldir returned %i, *pos 0x%llx, returning 0.\n",
-				rc, (long long)*pos);
-#endif
+	ntfs_log_debug("EOD, *pos 0x%llx, returning 0.\n", (long long)*pos);
 	return 0;
 dir_err_out:
 	errno = EIO;
@@ -1481,8 +1486,10 @@ search:
 	 * case there are no reference to this inode left, so we should free all
 	 * non-resident attributes and mark all MFT record as not in use.
 	 */
-	if (ni->mrec->link_count)
-		goto out;
+	if (ni->mrec->link_count) {
+		ntfs_inode_update_times(ni, NTFS_UPDATE_CTIME);
+		goto ok;
+	}
 	ntfs_attr_reinit_search_ctx(actx);
 	while (!ntfs_attrs_walk(actx)) {
 		if (actx->attr->non_resident) {
@@ -1523,13 +1530,17 @@ search:
 				"Leaving inconsistent metadata.\n");
 	}
 	ni = NULL;
+ok:	
+	ntfs_inode_update_times(dir_ni, NTFS_UPDATE_MCTIME);
 out:
 	if (actx)
 		ntfs_attr_put_search_ctx(actx);
 	if (ictx)
 		ntfs_index_ctx_put(ictx);
-	if (ni)
-		ntfs_inode_close(ni);
+	if (ntfs_inode_close(dir_ni) && !err)
+		err = errno;
+	if (ntfs_inode_close(ni) && !err)
+		err = errno;
 	if (err) {
 		errno = err;
 		ntfs_log_debug("Could not delete file: %s\n", strerror(errno));
